#include <assert.h>
#include <stdlib.h>
#include <iostream>
#include <sstream>

#include "programInfo.hh"
#include "testcase.hh"


int smallestPowerOfTwoAtLeast(int n)
{
  int i = 1;
  while (1)
  {
    if (i >= n)
      return i;
    i *= 2;
    assert(i != 0);
  }
}


// Should implicitly serialize, back up, log, cache.
int testTargetWithTestcase(programInfo &p, testcase &t)
{
  t.writeToDefault();
  return p.test();
}


// Use a binary-search-like algorithm to reduce the testcase.
// Usually O(m * lg n) in oracle calls, where m is the size of the minimal testcase and n is the size of the original testcase.
// Worst-case O(n^2).
// Postcondition: 1-minimal :)
void reduceTestcase(programInfo &p, testcase &t)
{
  int totalSize = t.parts.size();
  int chunkSize = smallestPowerOfTwoAtLeast(totalSize);
  if (chunkSize > 1)
    chunkSize = chunkSize / 2;
  
  int desiredReturnCode = testTargetWithTestcase(p, t); // this is typically a number representing "crash" or "hang".

  //assert(desiredReturnCode != 0); // e.g. success
  //assert(desiredReturnCode != 1); // e.g. can't find program

  cout << "Desired return code: " << desiredReturnCode << endl;
  cout << "Is that ok?" << endl;
  char temp;
  cin >> temp;
  assert(temp == 'y' || temp == 'Y');
  
  // Perhaps there should be an interactive "Is that desiredReturnCode correct?" prompt.
  // Or, as a sanity check, comment out the entire testcase and see if it changes.
  
  while (1)
  {
    cout << "Chunk size: " << chunkSize << endl;
  
    int chunksRemoved = 0;
    int chunksSurviving = 0;
    
    for (int chunkStart = 0; chunkStart < totalSize; chunkStart += chunkSize) {
      int chunkEnd = min(totalSize, chunkStart + chunkSize);
      
      assert(chunkEnd > chunkStart);
      
      // Is this chunk of the testcase already completely suppressed?  If so, don't bother.
      bool testThisChunk = false;
      for (int pos = chunkStart; pos < chunkEnd; ++pos) {
        if (t.parts[pos]->suppressed == 0) {
          testThisChunk = true;
          break;
        }
      }

      //if (!testThisChunk)
      //  cout << "Skipping chunk " << chunkStart << ".." << (chunkEnd - 1) << endl;
      
      if (testThisChunk) {
        // Suppress this entire chunk.
        for (int pos = chunkStart; pos < chunkEnd; ++pos) {
          ++ t.parts[pos]->suppressed;
        }
        
        if (testTargetWithTestcase(p, t) == desiredReturnCode) {
          cout << "Yay, reduced testcase by removing " << chunkStart << ".." << (chunkEnd - 1) << " :)" << endl;
          
          ++chunksRemoved;
        } else {
          cout << "Removing " << chunkStart << ".." << (chunkEnd - 1) << " made the problem go away.  Putting it back." << endl;

          ++chunksSurviving;

          // Revert (unsuppress this entire chunk).
          for (int pos = chunkStart; pos < chunkEnd; ++pos) {
            -- t.parts[pos]->suppressed;
          }
        }
      }

    }
    
    // Save it.
    ostringstream oss;
    oss << "didcs" << chunkSize << ".bak"; // this should be based on the testcase's filename somehow..
    string outfilename = oss.str();
    t.writeToFile(outfilename.c_str());
    
    cout << endl;
    cout << "Done with a round of chunk size " << chunkSize << "!" << endl;
    cout << chunksSurviving << " chunks survived; " << chunksRemoved << " were removed" << endl;
    cout << "I think that means the testcase now has " << (chunksSurviving * chunkSize) << " lines/whatevers left." << endl;
    cout << "Intermediate testcase saved as " << outfilename << endl;
    cout << endl;
        
    if (chunkSize == 1) {
      if (chunksRemoved == 0) {
        cout << "Done!" << endl;
        break; // We're completely done. The testcase is now 1-minimal.
      }
      // Otherwise, continue with the same chunk size (1) to see if we can eliminate any more lines on an extra pass.
    } else {
      // Always continue with the next smaller chunk size.
      chunkSize /= 2;
      assert(chunkSize != 0);
    }
    
    cout << "Continuing with chunk size " << chunkSize << endl;
  }
  
  // Print statistics: number of tests, original size, final size, time spent, a graph, whatever.


  t.writeToDefault();
}




int main(int argc, char **argv)
{
  if (argc < 3) {
    cerr << "Usage: ./lithium (TIY | L | C) testcaseFilename" << endl;
    return 1;
  }
  
  char *testcaseMode = argv[1];
  char *testcaseFilename = argv[2];

  char *args[] = 
    {
      "", // eh?
      "30", // timeout for hangs
      "-", // prefix for logs?
//      "/Users/admin/Desktop/builds/DeerPark 2006-02-22.app/Contents/MacOS/firefox-bin", // filename -- I use firefox-bin because "firefox" shell script is buggy.

      "/Users/admin/Desktop/builds/DeerPark 2006-03-05.app/Contents/MacOS/firefox-bin", // filename -- I use firefox-bin because "firefox" shell script is buggy.
      
//      "/Users/admin/trunk/mozilla/fx-debug-shared-gcc401/dist/DeerParkDebug.app/Contents/MacOS/firefox-bin",
      testcaseFilename, // args to firefox
      ((char *)0) // this list has to be null-terminated
    };

//  programInfo p("./timed_run.py", args);

//  char *args[] = { "", ((char *)0) }; // DOESN'T WORK. WHY THE FUCK NOT?
  // programInfo p("./hits_my_assertion.sh", args);
  programInfo p("./leak_domwindow.sh", args);

  testcase t;
  t.setParseModes(testcaseMode);
  t.readFromFile(testcaseFilename);
  //t.show();
  t.writeToFile("testcase-backup");
  
  assert(t.parts.size() >= 2);

  reduceTestcase(p, t);
}
