#include <iostream>
#include <fstream>
#include <assert.h>
#include "testcase.hh"
using namespace std;


void testcasePart::treeOutput(ostream &out, bool yankMode, string newIndentation) const
{
  // Use a new indentation scheme for output, so it looks clean even if some things were yanked.
  // Note that this modifies the output.
  
  /*
  string indentation = "";
  for (int i = 0; i < totalIndentation; ++i)
    indentation += " ";
  */
  
  bool showThis = (suppressed == 0);
  bool showChildren = showThis || yankMode;

  if (showThis)
    newIndentation += "    ";
  
  if (showThis)
    if (!str1.empty()) // e.g. root
      out << newIndentation << str1;
  
  if (showChildren)
    for (unsigned int i = 0; i < children.size(); ++i)
      children[i]->treeOutput(out, yankMode, newIndentation);
  
  if (showThis)
    if(!str2.empty()) // e.g. leaves; python
      out << newIndentation << str2;
}

void testcase::writeToDefault() const
{
  writeToFile(naturalFilename.c_str());
}

void testcase::writeToFile(const char *filename) const
{
  ofstream file;
  file.open(filename);

  writeToStream(file);

  file.close();
}

void testcase::writeToStream(ostream &out) const
{
  out << prefix;

  if (parseMode == TREE) {
    treeRoot->treeOutput(out, (treeReductionMode == YANK), "");
  } else {
    for (unsigned int i=0; i<parts.size(); ++i)
      if (parts[i]->suppressed == 0)
        out << parts[i]->str1;
  }

  out << postfix;
} 


void testcase::setParseModes(const string &parseInfo)
{
  if (parseInfo[0] == 'T') {
    parseMode = TREE;

    if (parseInfo[1] == 'N')
      treeParseMode = NO_CLOSING_LINES;
    else if (parseInfo[1] == 'I')
      treeParseMode = CLOSING_LINES_IFF_CHILDREN;
    else
      cerr << "Warning: didn't set treeParseMode" << endl;
      
    if (parseInfo[2] == 'Y')
      treeReductionMode = YANK;
    else if (parseInfo[2] == 'S')
      treeReductionMode = DELSUBTREE;
    else
      cerr << "Warning: didn't set treeReductionMode" << endl;
  }
  else if (parseInfo[0] == 'L')
    parseMode = LINES;
  else if (parseInfo[0] == 'C')
    parseMode = CHARS;
  else
    cerr << "Warning: didn't set parseMode" << endl;
}

void testcase::readFromFile(const char *filename)
{
  naturalFilename = filename;

  ifstream file;
  file.open(filename);
  
  if (!file.good())
    cerr << "Testcase file " << filename << " cannot be read." << endl;

  string line;

  /* 
   * Read the lines up to and including the one containing DDBEGIN into |prefix|, 
   * the lines up to the one containing "DDEND" into |parts| (line by line, including the line break character),
   * and the rest into |postfix|.
   */
  
  // Note: this function's behavior when DDBEGIN / DDEND are missing don't make a whole lot of sense.
  
  // Note: this function doesn't remember whether there was a newline at the end of the last line in the file, so the rewritten file may differ slightly.
  
  // Note: this function may assume unix line endings.

  while(getline(file, line) && line.find("DDBEGIN") == string::npos)
    prefix += line + "\n";
    
  prefix += line + "\n"; // DDBEGIN line is part of prefix.
  
  start();
  
  while(getline(file, line) && line.find("DDEND") == string::npos)
    lineAvailable(line + "\n");
  
  end();
    
  postfix += line + "\n"; // DDEND line is part of postfix.

  while(getline(file, line))
    postfix += line + "\n";
    
  file.close();
}


void testcase::start()
{
  if (parseMode == TREE) {
    testcasePart *root = new testcasePart("");
    root->totalIndentation = ROOT_INDENTATION;
    indentationStack.push_back(root);
    treeRoot = root;
  }
}

void testcase::lineAvailable(const string &line)
{

  if (parseMode == LINES) {

    parts.push_back(new testcasePart(line));

  } else if (parseMode == CHARS) {

    for (unsigned int i = 0; i < line.length(); ++i) {
      string t = line.substr(i, 1); // !
      parts.push_back(new testcasePart(t));
    }

  } else if (parseMode == TREE) {

    treeLineAvailable(line);

  }

}

void testcase::end()
{
  if (parseMode == TREE) {
    if (usesSpaces && usesTabs)
      cerr << "Warning: testcase uses both tabs and spaces." << endl;
    if (indentationStack.size() > 2) // :(
      cerr << "Warning: " << (indentationStack.size() - 2) << " unclosed things at end." << endl;
  }
}

testcase::~testcase()
{
  if (treeRoot)
    delete treeRoot;
  // Should also loop through parts and delete them.
}

void testcase::treeLineAvailable(const string &line)
{
  // Find this line's indentation level.
  unsigned int i = 0;
  int indentation = 0;
  unsigned int lineLength = line.length();
  for (i = 0; i < lineLength; ++i) {
    if (line[i] == ' ') {
      indentation += 1;
      usesSpaces = true;
    } else if (line[i] == '\t') {
      indentation += TAB_SPACE_COUNT;
      usesTabs = true;
    } else if (line[i] == '\n') {
      // Count this as whitespace for the purpose of determining whether the line is empty.
    } else {
      break;
    }
  }
  
  // Ignore empty lines.
  //cerr << i << " " << lineLength << endl;
  if (i == lineLength) {
    //cerr << "Ignoring line that is empty except for whitespace." << endl;
    return;
  }
    
  string lineWithoutIndentation = line.substr(i);
  
  // Figure out where on the tree it should go.
  // Invariant: |top| is something this could be a child, sibling, or closing-tag (in CLOSING_LINES_IFF_CHILDREN mode) for.
  // If the previous line was an opening tag, it's the previous line.
  // If the previous line was a closing tag (in CLOSING_LINES_IFF_CHILDREN mode), it's that tag's parent.
  
  // Should this be a loop??
  
  while(1) {
    testcasePart *top = indentationStack.back();
    assert(top);
    
    // There are three possibilities:
    // - We're a child of |top|.
    // - We're a closing tag for |top|. (in CLOSING_LINES_IFF_CHILDREN mode)
    // - We're a child of something above |top|.
    // Test for these in order.
    
    if (indentation > top->totalIndentation) {
      // We are a child of |top|.
      
      if (!top->children.empty())
        if (top->children.back()->totalIndentation != indentation)
          cerr << "Warning: I have an older sibling at a different indentation level!" << endl;

      // Create a testcasePart.  Add it to three lists (!).
      testcasePart *tp = new testcasePart(lineWithoutIndentation);
      tp->totalIndentation = indentation;
      tp->parent = top;

      parts.push_back(tp);
      top->children.push_back(tp);
      indentationStack.push_back(tp);

      break;
    }
    
    if (treeParseMode == CLOSING_LINES_IFF_CHILDREN && !top->children.empty()) {
      // We're a closing tag for |top|.
      // I think this can only happen on the second go through the loop (not the first, not the third or more).
      
      if (top->totalIndentation != indentation)
        cerr << "Warning: Closing tag not found or not at the expected indentation level!" << endl;
      
      top->str2 = lineWithoutIndentation;

      assert(top->totalIndentation != ROOT_INDENTATION);
      indentationStack.pop_back();
      assert(!indentationStack.empty());
      
      break;
    }
    
    // Otherwise, need to look higher.  Pop and continue through the loop.
    assert(top->totalIndentation != ROOT_INDENTATION);
    indentationStack.pop_back();
    assert(!indentationStack.empty());
  }

}
